/*
 * Copyright (c) 2002 by Louis Zechtzer
 *
 * Permission to use, copy and distribute this software is hereby granted
 * under the terms of version 2 or any later version of the GNU General Public
 * License, as published by the Free Software Foundation.
 *
 * THIS SOFTWARE IS PROVIDED IN ITS "AS IS" CONDITION, WITH NO WARRANTY
 * WHATSOEVER. NO LIABILITY OF ANY KIND FOR ANY DAMAGES WHATSOEVER RESULTING
 * FROM THE USE OF THIS SOFTWARE WILL BE ACCEPTED.
 */
/* 
 * Authors: Louis Zechtzer (lou@clarity.net)
 * Authors: G4SRA (cradford@bcs.org.uk)
 */

#include <stdio.h>
#include <stdlib.h>
#include <getopt.h>
#include <fcntl.h>
#include <errno.h>

#include "omd.h"
#include "tcp.h"
#include "net.h"
#include "sys.h"
#include "openmosix.h"
#include "vector.h"
#include "node.h"
#include "autodisc.h"
#include "bequest.h"
#include "omsyslog.h"
#include "omstatus.h"

static net_params_t net_handle;

/*
 * omd_message_handler: will be called whenever a multicast message is
 * received.  It practice it should handle adding new nodes to
 * the map.  The network functions pass (a potentially dangling reference) 
 * to a net_msg_t struct and "data" which is the net_params object.
 */
static void
omMessageHandler (net_msg_t * msg, void *data)
{
  switch (msg->type) {
  case NET_MSG_JOIN:
    /* Received a muticast join */
    omLog (OM_LOG_INFO, "Received JOIN from %s (%08x)",
	   inet_ntoa (msg->src), ntohl (msg->src.s_addr));
    /* Self-Connect to add the remote node into the OpenMosix map */
    /* If the node is already in the map, it will be ignored */
    tcp_send_selff (MSG_BEQ_NA_JOIN "%08x\n", msg->src.s_addr);
    /*
     * If routing is enabled, echo message out to all interfaces
     * except the one which received the message.
     */
    /* net_route_msg(&msg, (net_params_t *)data); */

    /* Notify new node (and others) via multicast, of this node's existence */
    net_send_ack_msg ((net_params_t *) data);
    break;

  case NET_MSG_ACK:
    /* Received a multicast ack (in responce to our/someone's multicast join) */
    omLog (OM_LOG_INFO, "Received ACK from %s (%08x)",
	   inet_ntoa (msg->src), ntohl (msg->src.s_addr));
    /* Self-Connect to add the remote node into the OpenMosix map */
    /* If the node is already in the map, it will be ignored */
    tcp_send_selff (MSG_BEQ_NA_JOIN "%08x\n", msg->src.s_addr);

    /*
     * If routing is enabled, echo the acknowledgment to all
     * interfaces except the one that received the message 
     */
    /* net_route_msg(&msg); */
    break;

  case NET_MSG_MASTER:
    /* Received a multicast master - request to do work */
    omLog (OM_LOG_INFO, "Received MASTER from %s (%08x)",
	   inet_ntoa (msg->src), ntohl (msg->src.s_addr));
    /* FIXME - DO SOMETHING ABOUT IT (or not :>) */
    break;
  case NET_MSG_DRONE:
    /* Received a multicast drone - advert free to do work */
    omLog (OM_LOG_INFO, "Received DRONE from %s (%08x)",
	   inet_ntoa (msg->src), ntohl (msg->src.s_addr));
    /* FIXME - DO SOMETHING ABOUT IT (or not :>) */
    break;
  case NET_MSG_GRID:
    /* Received a multicast node - advert free to ??? */
    omLog (OM_LOG_INFO, "Received GRID from %s (%08x)",
	   inet_ntoa (msg->src), ntohl (msg->src.s_addr));
    /* FIXME - DO SOMETHING ABOUT IT (or not :>) */
    break;
  case NET_MSG_EXIT:
    /* Received a multicast exit */
    omLog (OM_LOG_INFO, "Received EXIT from %s (%08x)",
	   inet_ntoa (msg->src), ntohl (msg->src.s_addr));
    /* Self-Connect to delete the remote node from the OpenMosix map */
    /* If the node is not in the map, openmosix_del_node will ignore it */
    tcp_send_selff (MSG_BEQ_NA_EXIT "%08x\n", msg->src.s_addr);
    break;
  default:
    omLog (OM_LOG_ALERT,
	   "Received unknown multicast type \"%x\" from %s (%08x)",
	   inet_ntoa (msg->src), msg->type, ntohl (msg->src.s_addr));
  }
}
int
omGotSigint (void)
{
  sys_sighandled &= ~SYS_GOT_SIGINT;
#ifdef _DEBUG_
  omLog (OM_LOG_DEBUG, "got SIGINT");
#endif
  if (!(sys_flags & SYS_FLAGS_DAEMON)) {
    omLog (OM_LOG_INFO, "UDP Server non-daemon exiting on SIGINT.");
    net_send_exit_msg (&net_handle);
    net_finalize (&net_handle);
    return OM_STATUS_CRIT;
  }
  /* reinitialise everything and restart afresh
     else
     {
     net_send_exit_msg (&net_handle);
     net_finalize (&net_handle);
     return OM_STATUS_RESTART;
     }
   */
  return OM_STATUS_SUCCESS;
}

int
omGotSigterm (void)
{
  sys_sighandled &= ~SYS_GOT_SIGTERM;
  omLog (OM_LOG_INFO, "UDP Server daemon exiting on SIGTERM.");
  net_send_exit_msg (&net_handle);
  net_finalize (&net_handle);
  return OM_STATUS_CRIT;
}

void
omGotSigalrm (void)
{
  sys_sighandled &= ~SYS_GOT_SIGALRM;
#ifdef _DEBUG_
  omLog (OM_LOG_DEBUG, "got SIGALRM");
#endif
  net_send_ack_msg (&net_handle);
}

void
omGotSighup (void)
{
  sys_sighandled &= ~SYS_GOT_SIGHUP;
#ifdef _DEBUG_
  omLog (OM_LOG_DEBUG, "got SIGHUP");
#endif
  omLogToggleDebug ();
}

void
omGotSigusr1 (void)
{
  sys_sighandled &= ~SYS_GOT_SIGUSR1;
#ifdef _DEBUG_
  omLog (OM_LOG_DEBUG, "got SIGUSR1");
#endif
  if (sys_flags & SYS_FLAGS_CONTROLLER) {
    /* Someone (not us) sent this */
    omLog (OM_LOG_WARNING, "Unexpected SIGUSR1 on Controller");
  }
  else if (sys_flags & SYS_FLAGS_DRONE) {
    /* Idle drone(+grid) - advertise our willingness to work */
    net_send_drone_msg (&net_handle);
  }
  else if (sys_flags & SYS_FLAGS_GRID) {
    /* FIXME: IDLE GRID NODE - advertise our willingness to work */
    net_send_grid_msg (&net_handle);
  }
  else {
    /* Someone (not us) sent this */
    omLog (OM_LOG_WARNING, "Unexpected SIGUSR1");
  }
}

void
omGotSigusr2 (void)
{
  sys_sighandled &= ~SYS_GOT_SIGUSR2;
#ifdef _DEBUG_
  omLog (OM_LOG_DEBUG, "got SIGUSR2");
#endif
  if (sys_flags & SYS_FLAGS_CONTROLLER) {
    /* Busy controller(+grid) - advertise our need for someone to do work */
    net_send_master_msg (&net_handle);
  }
  else if (sys_flags & SYS_FLAGS_DRONE) {
    /* Someone (not us) sent this */
    omLog (OM_LOG_WARNING, "Unexpected SIGUSR2 on Drone");
  }
  else if (sys_flags & SYS_FLAGS_GRID) {
    /* FIXME: BUSY GRID NODE - advertise our need for someone to do work */
    net_send_grid_msg (&net_handle);
  }
  else {
    /* Someone (not us) sent this */
    omLog (OM_LOG_WARNING, "Unexpected SIGUSR2");
    net_send_join_msg (&net_handle);
  }
}

int
omAutoDiscovery (int opt_mcast_ttl, char *if_names)
{
  struct in_addr if_addrs[NET_MAX_IFS];
  int cnt;
  int ret;

  /* Initialize network transport */
  if (net_initialize (&net_handle, if_names, opt_mcast_ttl) == NET_FAIL) {
    omLog (OM_LOG_CRIT, "Unable to initialize network. Exiting.");
    return OM_STATUS_CRIT;
  }

  /* 
   * Initialize openMosix specifics.  Initial entries to the kernel's
   * map should include one node-id and interface, and aliases for    
   * each additional interface.
   */
  net_get_if_addrs (&net_handle, if_addrs, &cnt);
  /*
   * Send a multicast join message, this will cause the creation of
   * our own node entry into the tcp OpenMosix kernel map,
   * NO IT WONT, WE DO NOT PROCESS OUR OWN MULTICASTS - G4SRA
   * and in addition, add us to any other listening nodes.
   */
  net_send_join_msg (&net_handle);
  /*
   * Pass all received multicast messages to the message handler,
   * stopping only to process signals .
   */
  for (;;) {
    net_recv_msg (&net_handle, &omMessageHandler, (void *) &net_handle);
    if (sys_sighandled) {
      if (sys_sighandled & SYS_GOT_SIGINT) {
	if ((ret = omGotSigint ()) < 0)
	  return (ret);
      }
      if (sys_sighandled & SYS_GOT_SIGTERM) {
	if ((ret = omGotSigterm ()) < 0)
	  return (ret);
      }
      if (sys_sighandled & SYS_GOT_SIGALRM) {
	omGotSigalrm ();
      }
      if (sys_sighandled & SYS_GOT_SIGHUP) {
	omGotSighup ();
      }
      if (sys_sighandled & SYS_GOT_SIGUSR1) {
	omGotSigusr1 ();
      }
      if (sys_sighandled & SYS_GOT_SIGUSR2) {
	omGotSigusr2 ();
      }
      /* NOTE: inconclusive, we could arrive here via a logic race */
      if (sys_sighandled) {
	omLog (OM_LOG_NOTICE, "Unhandled Signal %x", sys_sighandled);
	sys_sighandled = 0;
      }
    }
  }
  omLog (OM_LOG_ALERT, "the impossible happened! :%s", strerror (errno));
  return OM_STATUS_FAIL;
}
