/*
 * Copyright (c) 2002 by Louis Zechtzer
 * Copyright (c) 2005 by Nathan Wiegand
 *
 * Permission to use, copy and distribute this software is hereby granted
 * under the terms of version 2 or any later version of the GNU General Public
 * License, as published by the Free Software Foundation.
 *
 * THIS SOFTWARE IS PROVIDED IN ITS "AS IS" CONDITION, WITH NO WARRANTY
 * WHATSOEVER. NO LIABILITY OF ANY KIND FOR ANY DAMAGES WHATSOEVER RESULTING
 * FROM THE USE OF THIS SOFTWARE WILL BE ACCEPTED.
 */
/* 
 * Authors: Louis Zechtzer (lou@clarity.net)
 * Authors: Nathan Wiegand (nathanwiegand@gmail.com)
 * Authors: G4SRA (cradford@bcs.org.uk)
 */

#include <stdio.h>
#include <stdlib.h>
#include <getopt.h>
#include <fcntl.h>
#include <errno.h>
#include <netdb.h>
#include <sys/types.h>
#include <sys/wait.h>
#include <string.h>
#include <libgen.h>

#include "omd.h"
#include "tcp.h"
#include "net.h"
#include "omsyslog.h"
#include "omstatus.h"
#include "sys.h"
#include "vector.h"
#include "node.h"
#include "omlocal.h"
#include "autodisc.h"
#include "openmosix.h"

char *program_name = NULL;

int omd_parent_pid;
int omd_child_pid;
int omd_this_pid;

int opt_loglevel = 0;

char *if_names = NULL;
int mcast_ttl = 0;

static void
usage ()
{
  fprintf (stderr, "Usage: %s\n"
	   "\t\t[ -i|--interface <interface_name> ]\n"
	   "\t\t[ -t|--tcp <tcp port> ] [ -u|--udp <udp port> ]\n"
	   "\t\t[ -m|--multicast <ip> ] [ -o|--multicast_ttl <#> ]\n"
	   "\t\t[ -n|--nodaemon ] [ -l|--loglevel <#> ]\n"
	   "\t\t[ -p|--poll <#> ] [ -a|--age <#> ]\n"
	   "\t\t[ [ -c|--controller ] || [ -d|--drone ] ]\n"
	   "\t\t[ -g|--grid ] [ -q|--qfactor ] [ -b|buggeritup ]\n"
	   "\t\t|| [ -V|--version ]\n"
	   "\t\t|| [ -?|--usage ] [ -h|--help ]\n\n", program_name);
}

void
help (void)
{
  fprintf (stderr, "\n"
	   "\t\tloglevel (syslog(3) bitmask) 255 is everything\n"
	   "\t\tpoll activity monitor interval in seconds\n"
	   "\t\tage max seconds to retain volatile node information\n"
	   "\n"
	   "\t\tcontroller automatically migrates workload to drones\n"
	   "\t\tqfactor is level above which to migrate workload\n"
	   "\t\tgrid multicast request for any node to do work\n"
	   "\n"
	   "\t\tdrone automatically accept workload from controller\n"
	   "\t\tqfactor level below which to request work\n"
	   "\t\tgrid multicast request for any node to assign work\n" "\n");
}

static void
omd_version ()
{
  fprintf (stderr,
	   "%s (OpenMosix Tools) " "OpenMosix Userspace Daemon "
	   OMD_VERSION "\n" "Copyright (C) By numerous contributors\n"
	   "This is free software; see the source for copying conditions.  There is NO\n"
	   "warranty; not even for MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.\n",
	   program_name);
}

int
omd_commandline_args (int argc, char *argv[])
{

  static struct option long_opts[] = {
    {"age", 1, NULL, 'a'},
    {"buggeritup", 0, NULL, 'b'},
    {"controller", 0, NULL, 'c'},
    {"drone", 0, NULL, 'd'},
    {"grid", 0, NULL, 'g'},
    {"help", 0, NULL, 'h'},
    {"interface", 1, NULL, 'i'},
    {"loglevel", 1, NULL, 'l'},
    {"multicast", 1, NULL, 'm'},
    {"nodaemon", 0, NULL, 'n'},
    {"multicast_ttl", 1, NULL, 'o'},
    {"poll", 1, NULL, 'p'},
    {"qfactor", 1, NULL, 'q'},
    {"tcp", 1, NULL, 't'},
    {"udp", 1, NULL, 'u'},
    {"version", 0, NULL, 'V'},
    {"usage", 0, NULL, '?'},
    {0, 0, 0, 0}
  };

  int opt;
  int option_index;


  for (;;) {
    opt =
      getopt_long (argc, argv, "a:bcdghi:l:m:no:p:q:t:u:V?", long_opts,
		   &option_index);
    if (opt == -1) {
      break;
    }
    switch (opt) {
    case 'a':
      omd_node_life = atoi (optarg);
      break;
    case 'b':
      sys_flags |= SYS_FLAGS_POWER;
      break;
    case 'c':
      sys_flags |= SYS_FLAGS_CONTROLLER;
      break;
    case 'd':
      sys_flags |= SYS_FLAGS_DRONE;
      break;
    case 'g':
      sys_flags |= SYS_FLAGS_GRID;
      break;
    case 'i':
      if_names = optarg;
      break;
    case 'l':
      opt_loglevel = atoi (optarg);
      break;
    case 'm':
      net_mcast_addr = optarg;
      break;
    case 'n':
      sys_flags &= ~SYS_FLAGS_DAEMON;
      break;
    case 'o':
      mcast_ttl = atoi (optarg);
      break;
    case 'p':
      omd_poll = atoi (optarg);
      break;
    case 'q':
      sys_qfactor = atof (optarg);
      break;
    case 't':
      mosix_tcp_port_nbo = htons (atoi (optarg));
      break;
    case 'u':
      net_mcast_port_nbo = htons (atoi (optarg));
      break;
    case 'V':
      omd_version ();
      exit (0);
    case 'h':
      usage ();
      help ();
      return (1);
    case '?':
      usage ();
      exit (0);
      break;
    };
  }

  if (sys_flags & (SYS_FLAGS_CONTROLLER | SYS_FLAGS_DRONE | SYS_FLAGS_GRID)) {
    if (openmosix_check () != 0) {
      usage ();
      fprintf (stderr,
	       "\n\tOptions [ -c|--controller ], [ -d|--drone ], or [ -g|--grid ]\n"
	       "\t\t not an OpenMosix Kernel\n\n");
      return (1);
    }
  }

  if ((sys_flags & SYS_FLAGS_CONTROLLER) && (sys_flags & SYS_FLAGS_DRONE)) {
    usage ();
    fprintf (stderr,
	     "\n\tOptions [ -c|--controller ] and [ -d|--drone ] are mutually exclusive.\n\n");
    return (1);
  }

  if (sys_qfactor != 0) {
    if (!
	(sys_flags &
	 (SYS_FLAGS_CONTROLLER | SYS_FLAGS_DRONE | SYS_FLAGS_GRID))) {
      usage ();
      fprintf (stderr,
	       "\n\tOption [ -q|qfactor <#> ] is meaningless without one of\n"
	       "\t\t[ -c|--controller ] or [ -d|--drone ] or [ -g|--grid ]\n\n");
      return (1);
    }
    /* FIXME: Rangecheck for sensible values here */
  }
  else {
    if (sys_flags & SYS_FLAGS_CONTROLLER) {
      /* trigger migrate when average load exceeds */
      sys_qfactor = 1.0;
    }
    else if (sys_flags & SYS_FLAGS_DRONE) {
      /* trigger idle when average load is less */
      sys_qfactor = 0.25;
    }
    else if (sys_flags & SYS_FLAGS_GRID) {
      /* trigger migrate/idle when average load exceeds/less */
      sys_qfactor = 0.5;
      /* idle/2=0.25 busy*2=1.0 */
    }
  }
  return (0);
}

int
main (int argc, char *argv[])
{
  char *tmpstr;


  program_name = omStrdup (basename (argv[0]));

  /* Maybe sometime in the future, but not now */
  /*setuid (getuid ()); */

  if (omd_commandline_args (argc, argv))
    exit (-1);

  /* Initialise logging facility */
  /* FIXME: merge opt_loglevel into omLogInitialise */
  omLogInitialise (sys_flags & SYS_FLAGS_DAEMON ? 0 : 1, program_name);
  if (opt_loglevel != (0))
    omLogSetMask (opt_loglevel);

  if (mosix_tcp_port_nbo == 0) {
    struct servent *p;
    if ((p = getservbyname (program_name, "tcp")) != NULL)
      mosix_tcp_port_nbo = (unsigned short) p->s_port;
    else {
      omLog (OM_LOG_ALERT, "Service \"%s\" no tcp port defined (%d/tcp).",
	     program_name, MOSIX_TCP_PORT);
      exit (-1);
    }
  }

  if (net_mcast_port_nbo == 0) {
    struct servent *p;
    if ((p = getservbyname (program_name, "udp")) != NULL)
      net_mcast_port_nbo = (unsigned short) p->s_port;
    else {
      omLog (OM_LOG_ALERT, "Service \"%s\" no udp port defined (%d/udp).",
	     program_name, NET_MCAST_PORT);
      exit (-1);
    }
  }

/*
  omLog (OM_LOG_NOTICE, "Service \"%s\" using ports %d/udp %d/tcp",
	   program_name, ntohs (net_mcast_port_nbo),
	   ntohs (mosix_tcp_port_nbo));
*/

  /* Initialize system aspects */
  omSysInitialize ();

  if ((omd_child_pid = fork ()) == (-1)) {
    omLog (OM_LOG_ALERT, "%s Failed to fork:%s", program_name,
	   strerror (errno));
    exit (-1);
  }
  else if (omd_child_pid == 0) {
    omd_parent_pid = getppid ();
    omd_this_pid = getpid ();
    /* Do UDP stuff */
    tmpstr = omMalloc (strlen (program_name) + sizeof ("(udp)"));
    strcpy (tmpstr, program_name);
    strcat (tmpstr, "(udp)");
    free (program_name);
    program_name = tmpstr;

    /* FIXME: merge opt_loglevel into omLogInitialise */
    omLogInitialise (sys_flags & SYS_FLAGS_DAEMON ? 0 : 1, program_name);
    if (opt_loglevel != (0))
      omLogSetMask (opt_loglevel);

    omLog (OM_LOG_NOTICE, "\"%s\" port %d/udp pid:%d ppid:%d started",
	   program_name, ntohs (net_mcast_port_nbo), omd_this_pid,
	   omd_parent_pid);
    usleep (500000);
    /* start the udp server timer interrupt - every 11 minuites */
    omSysPollInterval (60 * 11);
    if (omAutoDiscovery (mcast_ttl, if_names) == OM_STATUS_CRIT) {
      omLog (OM_LOG_NOTICE, "\"%s\" pid:%d ppid:%d exiting",
	     program_name, omd_this_pid, omd_parent_pid);
      if (getppid () > 1) {
	omLog (OM_LOG_INFO, "\"%s\" parent is alive, simple death",
	       program_name);
      }
      else {
	omLog (OM_LOG_INFO, "\"%s\" parent is dead, complex death",
	       program_name);
	/* Ill do this after the interesting stuff is done :>) */
      }
      omLogFinalise ();
      exit (0);
    }
  }
  else {
    uint32_t addr;
    int sfd;
    omd_parent_pid = getppid ();
    omd_this_pid = getpid ();
    /* Do TCP stuff */
    omLogFinalise ();
    tmpstr = omMalloc (strlen (program_name) + sizeof ("(tcp)"));
    strcpy (tmpstr, program_name);
    strcat (tmpstr, "(tcp)");
    free (program_name);
    program_name = tmpstr;
    /* FIXME: merge opt_loglevel into omLogInitialise */
    omLogInitialise (sys_flags & SYS_FLAGS_DAEMON ? 0 : 1, program_name);
    if (opt_loglevel != (0))
      omLogSetMask (opt_loglevel);
    omLog (OM_LOG_NOTICE, "\"%s\" port %d/tcp pid:%d cpid:%d started",
	   program_name, ntohs (mosix_tcp_port_nbo), omd_this_pid,
	   omd_child_pid);
    /* start the tcp server timer interrupt */
    omSysPollInterval (omd_poll);
    /* FIXME: - if our (udp) child dies we are only half a daemon */
    signal (SIGCHLD, SIG_IGN);
    /* now I don't have to wait()! */

    omd_tcp_server_init (&addr, &sfd);
    omd_initialise (addr);
    if (omd_tcp_server_accept (sfd) == OM_STATUS_CRIT) {
      kill (omd_child_pid, SIGTERM);
      omSleep (1);
/*
         if (waitpid (omd_child_pid, NULL, 0) == -1) {
         sys_errno = errno;
         if (sys_errno != ECHILD)
         omLog (OM_LOG_CRIT, "Problem TERMinating child: %d", omd_child_pid);
         }
*/
      omd_finalize ();
      omLog (OM_LOG_NOTICE, "\"%s\" pid:%d cpid:%d exiting",
	     program_name, omd_this_pid, omd_child_pid);
      omLogFinalise ();
      exit (0);
    }
  }
  omLog (OM_LOG_ALERT,
	 "%s pid:%d fell over, parent:%d, child:%d\nperror:%s", program_name,
	 omd_this_pid, omd_parent_pid, omd_child_pid, strerror (errno));
  exit (-1);
}
