/*
 * Copyright (c) 2002 by Louis Zechtzer
 * Copyright (c) 2005 by Nathan Wiegand 
 *
 * Permission to use, copy and distribute this software is hereby granted
 * under the terms of version 2 or any later version of the GNU General Public
 * License, as published by the Free Software Foundation.
 *
 * THIS SOFTWARE IS PROVIDED IN ITS "AS IS" CONDITION, WITH NO WARRANTY
 * WHATSOEVER. NO LIABILITY OF ANY KIND FOR ANY DAMAGES WHATSOEVER RESULTING
 * FROM THE USE OF THIS SOFTWARE WILL BE ACCEPTED.
 */
/* 
 * Authors: Louis Zechtzer (lou@clarity.net)
 * Authors: Nathan Wiegand (nathanwiegand@gmail.com)
 * Authors: G4SRA (cradford@bcs.org.uk)
 */

#include <sys/socket.h>
#include <netinet/in.h>
#include <sys/types.h>
#include <arpa/inet.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <string.h>
#include <stdint.h>
#include <unistd.h>
#include <dirent.h>

#include "request.h"
#include "bequest.h"
#include "vector.h"
#include "openmosix.h"
#include "omsyslog.h"
#include "omstatus.h"
#include "tcp.h"
#include "net.h"
#include "sys.h"
#include "semaphore.h"
#include "misc.h"
#include "omd.h"
#include "omwakeup.h"
#include "ommacaddr.h"
#include "telinit.h"
#include "misc.h"


vector *cluster;
int omd_node_life = SYS_NODE_LIFE;
int omd_poll = SYS_POLL_FREQ;

/*
Return the number of entries in the tcp server daemon OpenMosix node table.
*/
int
omd_cluster_size ()
{
  return cluster->end;
}

void
omd_age_node (omd_node * n)
{
  /* FIXME: this is the wrong place for this */
  if (omd_poll >= n->lifetime)
    omLog (OM_LOG_WARNING, "Node cache defunct (age - poll = %d)",
	   n->lifetime - omd_poll);
  if ((n->life -= omd_poll) <= 0) {
    /* Dont leave life < 0 */
    if (n->life < 0) {
      n->life = 0;
    }
    n->load_avg = (-1);
    n->free_mem = (-1);
  }
}

/* If we wish, we could automatically update our node as soon as it expires
 * else we can wait for a request for data and update in realtime
 */

void
omController (omd_node * n)
{

  /* controller - update node as soon as it expires */
  omNodeUpdate (n);
  /* FIXME: what about memory */
  if (n->load_avg > sys_qfactor) {
    pid_t pid = 0;
    /* Overworked controller - farm something out */
    omLog (OM_LOG_NOTICE, "Controller overloaded (%lf:%lf)",
	   n->load_avg, sys_qfactor);
    /* find the most suitable process */
    if ((pid = omCandidateProcess ()) > 0) {
      int nn;
      /* find the drone node with the least load */
      /* nn = omCandidateDrone (MAXDOUBLE); */
      /* find a drone with a load less than ours */
      nn = omCandidateDrone (n->load_avg);
      /* FIXME: the above choice should be switched by a flag */
      if (nn > 0) {
	/* farm out the process to the node */
	if (omMigrate (pid, nn) < OM_STATUS_SUCCESS) {
	  omLog (OM_LOG_NOTICE, "Migrating pid %d to node %d failed",
		 pid, nn);
	}
	else {
	  omLog (OM_LOG_NOTICE, "Migrated pid %d to node %d", pid, nn);
	  /* Cache needs update after migrate, make sure we do */
	  n->life = 0;
	}
      }
      else {
	omd_node *sn = NULL;
	int sent_wakeup = 0;

	if (n->flags & SYS_FLAGS_POWER) {
	  int i;

	  for (i = cluster->end - 1; i > 0; i--) {
	    /* Select any powered down (sleeping) nodes */
	    if ((sn = v_get (cluster, i))->life < 0) {
	      if (sn->flags & SYS_FLAGS_POWER) {
		if (omCompatible (i)) {
		  struct in_addr saddr;
		  saddr.s_addr = n->s_addr;
		  if (omWakeupNode (sn) >= OM_STATUS_SUCCESS) {
		    omLog (OM_LOG_NOTICE, "Waking node %s (%8x)",
			   inet_ntoa (saddr), ntohl (n->s_addr));
		    /* NOTE: sending the wol ok does not mean the node is/will awake */
		    ++sent_wakeup;
		    /*FIXME: If the node fails to wake, should it be deleted ? */

		    /* How many should we wake up ?, just one, two, or all ? */
		    /*
		       break;
		     */
		  }
		  else {
		    /*FIXME: If the send wol fails, should the node be deleted ? */
		    omLog (OM_LOG_NOTICE,
			   "Attempt to wake node %s (%8x) failed",
			   inet_ntoa (saddr), ntohl (n->s_addr));
		  }
		}
	      }
	    }
	  }
	}
	if (sent_wakeup == 0) {
	  if (n->flags & SYS_FLAGS_GRID) {
	    omLog (OM_LOG_NOTICE, "Advising need to migrate work");
	    /* Overworked controller - advertise our need (a multicast request) to migrate work */
	    kill (omd_child_pid, SIGUSR2);
	  }
	  else {
	    omLog (OM_LOG_NOTICE,
		   "No node for migratory suitable pid %d", pid);
	  }
	}
      }
    }
    else {
      omLog (OM_LOG_NOTICE, "No migratory suitable processes");
    }
  }
}

void
omDrone (omd_node * n)
{
  static int counting = 0;

  omNodeUpdate (n);
  if (n->load_avg < sys_qfactor) {
    if (n->flags & SYS_FLAGS_POWER) {
      counting += omd_poll;
      if (counting > (60 * 15)) {
	/* We have been idle for over fifteen minuites */
	/* FIXME: the powerdown runlevel should be user configurable */
	telinit ((int *) "0\0\0\0");
      }
    }
    else if (n->flags & SYS_FLAGS_GRID) {
      /* Idle drone - advertise our willingness to work */
      kill (omd_child_pid, SIGUSR1);
    }
  }
  else {
    if (n->flags & SYS_FLAGS_POWER) {
      counting = 0;
    }
  }
}

void
omGrid (omd_node * n)
{
  omNodeUpdate (n);
  if (n->load_avg < (sys_qfactor / 2)) {
    /* Idle node - advertise our willingness to work */
    kill (omd_child_pid, SIGUSR1);
  }
  else if (n->load_avg > (sys_qfactor * 2)) {
    /* Overworked node - advertise our need to migrate work */
    kill (omd_child_pid, SIGUSR2);
  }
}

void
omAgeNodeData (void)
{
  int i;
  omd_node *n;

  for (i = cluster->end - 1; i > 0; i--) {
    /* Ignore any zero (dormant) nodes */
    /* Ignore any negative (powered down) nodes */
    if ((n = v_get (cluster, i))->life > 0) {
      omd_age_node (n);
    }
  }

  /* Not ready for this just yet :>)
     if ((n = v_get (cluster, 0))->life >= 0) */  {
    n = v_get (cluster, 0);
    if (n->flags & SYS_FLAGS_CONTROLLER) {
      omController (n);
    }
    else if (n->flags & SYS_FLAGS_DRONE) {
      omDrone (n);
    }
    else if (n->flags & SYS_FLAGS_GRID) {
      omGrid (n);
    }
    else {
      omd_age_node (n);
    }
  }
}

/* Think of a universally acceptable (version portable) strategy */
void
omUpdateNodes (char *buf)
{
  int i;
  omd_node *n;

  for (i = cluster->end - 1; i > 0; i--) {
    /* Ignore any zero (dormant) nodes */
    /* Ignore any negative (powered down) nodes */
    if ((n = v_get (cluster, i))->life > 0) {
      ;

      /* send the remote node a message stating that
         the data it holds for us is incorrect */

      /* the remote node responds in its own time re-requesting
         all the information on us it is holding */
    }
  }
}

/*
Initialise the tcp server daemon OpenMosix node table.
*/
/* FIXME: Add ability to handle multiple interfaces */
int
omd_initialise (uint32_t addr)
/*omd_initialise (const struct in_addr *node_addr, int sz)*/
{
  omd_node *n;
  vector *vtmp;


  init_rwsem (&cluster_semaphore);

  down_write (&cluster_semaphore);
  if (cluster == NULL) {
    if ((cluster = create (4)) != NULL) {
      if ((n = omNodeCreate (addr)) != NULL) {
	if ((vtmp = v_add (cluster, n)) != NULL) {
	  cluster = vtmp;
	  omNodeInit (n);
	  n->lifetime = omd_node_life;
	  n->flags = sys_flags;
	  omNodeUpdate (n);
	  up_write (&cluster_semaphore);
	  {
	    struct in_addr saddr;
	    saddr.s_addr = addr;
	    omLog (OM_LOG_NOTICE, "Created %s as OpenMosix node %d",
		   inet_ntoa (saddr), cluster->end - 1);
	  }
	  return OM_STATUS_SUCCESS;
	}
      }
    }
    omLog (OM_LOG_CRIT, "OpenMosix: intialise out of memory");
  }
  else {
    omLog (OM_LOG_NOTICE, "OpenMosix: already initialised");
  }
  up_write (&cluster_semaphore);
  return OM_STATUS_FAIL;
}

/*
Decommission the tcp server daemon OpenMosix node table.
*/
int
omd_finalize ()
{
  uint32_t t_addr;
  vector *vtmp;
  omd_node *p;

  down_write (&cluster_semaphore);
  while (cluster->end > 0) {
    t_addr = (uint32_t) (v_get (cluster, cluster->end - 1))->s_addr;
    p = v_get (cluster, cluster->end - 1);
    omNodeDestroy (&p);
    if ((vtmp = v_remove (cluster, cluster->end - 1)) != NULL) {
      cluster = vtmp;
      omLog (OM_LOG_NOTICE, "Removed OpenMosix node %d (%08x)",
	     cluster->end, ntohl (t_addr));
    }
    else {
      up_write (&cluster_semaphore);
      omLog (OM_LOG_INFO, "Remove OpenMosix node %d (%08x) failed",
	     cluster->end - 1, ntohl (t_addr));
      return OM_STATUS_FAIL;
    }
  }
  up_write (&cluster_semaphore);
  return OM_STATUS_SUCCESS;
}

/*
Insert (install) a node entry into the tcp server daemon OpenMosix node table.
*/
/* Calling function must lock & unlock semaphore */
int
omd_ins_node (const struct in_addr *node_addr, const struct in_addr
	      *alias, int alias_cnt)
{
  omd_node *n;
  uint32_t t_addr = node_addr->s_addr;
  vector *vtmp;
  int i;

#ifdef _ROBUST_
  for (i = 0; i < cluster->end; i++) {
    if ((uint32_t) v_get (cluster, i)->s_addr == t_addr) {
      omLog (OM_LOG_ALERT, "ROBUST omd_ins_node trying to re-add %s (%08x)",
	     inet_ntoa (*node_addr), ntohl (t_addr));
      return OM_STATUS_NOACTION;
    }
  }
#endif

  /* Create the new (remote) node dataspace */
  if ((n = omNodeCreate (t_addr)) != NULL) {
    if (alias)
      n->alias = alias->s_addr;
    n->alias_cnt = alias_cnt;

    if ((vtmp = v_add (cluster, n)) != NULL) {
      cluster = vtmp;
      /* ++(v_get (cluster, 0)->num_nodes); */
      omLog (OM_LOG_NOTICE, "Added %s as OpenMosix node %d",
	     inet_ntoa (*node_addr), cluster->end - 1);
      omd_complete_node (node_addr);
      return OM_STATUS_SUCCESS;
    }
  }
  omLog (OM_LOG_INFO, "Add OpenMosix node %08x failed", ntohl (t_addr));
  return OM_STATUS_FAIL;
}

/*
Delete (remove) a node entry from the tcp server daemon OpenMosix node table.
*/
/* Calling function must lock & unlock semaphore */
int
omd_del_node (const struct in_addr *node_addr, const struct in_addr
	      *alias, int alias_cnt)
{
  int i;
  uint32_t t_addr = node_addr->s_addr;
  vector *vtmp;
  omd_node *p;

  /* This currently has no regard for aliases */
  for (i = 0; i < cluster->end; i++) {
    if ((uint32_t) v_get (cluster, i)->s_addr == t_addr) {
      p = v_get (cluster, i);
      omNodeDestroy (&p);
      if ((vtmp = v_remove (cluster, i)) != NULL) {
	cluster = vtmp;
	/* --(v_get (cluster, 0)->num_nodes); */
	omLog (OM_LOG_NOTICE, "Removed %s  OpenMosix node %d",
	       inet_ntoa (*node_addr), i);
	omd_print_cluster (OM_LOG_INFO);
	return OM_STATUS_SUCCESS;
      }
      omLog (OM_LOG_INFO, "Remove OpenMosix node %08x failed",
	     ntohl (t_addr));
      return OM_STATUS_FAIL;
    }
  }
  omLog (OM_LOG_INFO, "Remove OpenMosix node %08x, not found",
	 ntohl (t_addr));
  return OM_STATUS_NOACTION;
}

/*
Display the tcp server daemon OpenMosix node table.
*/
void
omd_print_cluster (int level)
{
  int i;
  char *ptr;

  down_read (&cluster_semaphore);
  if (cluster != NULL) {
    omLog (level, "Cluster size: %d", cluster->end);
    /* FIXME: sort out tidy column display of variable length fields */
    omLog (level, "|Nod|  Addr  |OMV|Arch|CPUG|Net|AvgLd|Dur|Age|Flag|");
    for (i = 0; i < cluster->end; i++) {
      omLog (level, "|%03d|%08x|%s|%s|%04d|%03d|%0+3.02lf|%03d|%03d|%04X|",
	     i,
	     ntohl (v_get (cluster, i)->s_addr),
	     (((ptr =
		get_node_omver (v_get (cluster, i))) != NULL) ? ptr : "---"),
	     (v_get (cluster, i)->arch !=
	      NULL ? v_get (cluster, i)->arch : "----"),
	     /*
	        (v_get (cluster, i)->num_cpus >
	        0 ? v_get (cluster, i)->num_cpus : 0), (v_get (cluster,
	        i)->mips !=
	        NULL ? v_get (cluster,
	        i)->
	        mips[0] : 0),
	      */
	     v_get (cluster, i)->grunt,
	     v_get (cluster, i)->net_cost,
	     v_get (cluster, i)->load_avg,
	     v_get (cluster, i)->lifetime, v_get (cluster, i)->life,
	     /*omMACAddrStr (v_get (cluster, i)->macaddr), */
	     (short) v_get (cluster, i)->flags);
    }
  }
  up_read (&cluster_semaphore);
}
