/*
 * Copyright (c) 2002 by Louis Zechtzer
 *
 * Permission to use, copy and distribute this software is hereby granted
 * under the terms of version 2 or any later version of the GNU General Public
 * License, as published by the Free Software Foundation.
 *
 * THIS SOFTWARE IS PROVIDED IN ITS "AS IS" CONDITION, WITH NO WARRANTY
 * WHATSOEVER. NO LIABILITY OF ANY KIND FOR ANY DAMAGES WHATSOEVER RESULTING
 * FROM THE USE OF THIS SOFTWARE WILL BE ACCEPTED.
 * 
 * Author: Louis Zechtzer (lou@clarity.net)
 * Author: G4SRA (cradford@bcs.org.uk)
 */

#include <stdio.h>
#include <stdlib.h>
#include <errno.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/time.h>
#include <signal.h>
#include <unistd.h>
#include <string.h>
#include <time.h>

#include "sys.h"
#include "omsyslog.h"
#include "omstatus.h"


double sys_qfactor = 0;
int sys_flags = SYS_FLAGS_DAEMON;
int sys_sighandled = 0;

int
omSysSigSetup (void (*handler) (int), int signal, int flags)
{
  struct sigaction sa;

  sigemptyset (&sa.sa_mask);
  sa.sa_handler = handler;
  sa.sa_flags = flags;
  return (sigaction (signal, &sa, NULL));
}

static void
omSysAlarmHandler (int s)
{
  sys_sighandled |= SYS_GOT_SIGALRM;
}

static void
omSysSignalHandler (int s)
{
  switch (s) {
  case SIGALRM:
    sys_sighandled |= SYS_GOT_SIGALRM;
    break;
  case SIGHUP:
    sys_sighandled |= SYS_GOT_SIGHUP;
    break;
  case SIGINT:
    sys_sighandled |= SYS_GOT_SIGINT;
    break;
  case SIGTERM:
    sys_sighandled |= SYS_GOT_SIGTERM;
    break;
  case SIGUSR1:
    sys_sighandled |= SYS_GOT_SIGUSR1;
    break;
  case SIGUSR2:
    sys_sighandled |= SYS_GOT_SIGUSR2;
    break;
  case SIGCHLD:
    sys_sighandled |= SYS_GOT_SIGCHLD;
    break;
  case SIGTSTP:
    sys_sighandled |= SYS_GOT_SIGTSTP;
    break;
  case SIGCONT:
    sys_sighandled |= SYS_GOT_SIGCONT;
    break;
  default:
    sys_sighandled |= SYS_GOT_SIGUNHANDLED;
    break;
  }
}

/* poll interval timer */
void
omSysPollInterval (int sec)
{
  struct itimerval pit;
  struct itimerval old_timer;

  pit.it_interval.tv_sec = sec;
  pit.it_interval.tv_usec = 0;
  pit.it_value.tv_sec = sec;
  pit.it_value.tv_usec = 0;
  if (setitimer (ITIMER_REAL, &pit, &old_timer) == 0) {
    omSysSigSetup (omSysAlarmHandler, SIGALRM, 0);
  }
  else {
    sys_errno = errno;
    omLog (OM_LOG_ALERT, "omSysPollInterval:%s", strerror (sys_errno));
  }
}

/*
 * sys_initialize: called to initialize the system module.  The system
 * module handles os specific things like daemons and signal handling.
 */
void
omSysInitialize (void)
{
  /* initialize signal handlers */
  struct sigaction sa;

  /* convert process to daemon if requested */
  if (sys_flags & SYS_FLAGS_DAEMON) {
    pid_t pid;

    /* Prepare for a guaranteed setsid() */
    if ((pid = fork ()) == -1) {
      omLog (OM_LOG_CRIT, "fork before setsid");
      exit (1);
    }
    else if (pid != 0) {
      exit (0);
    }

    /* Set up the new process group with us as leader */
    setsid ();

    if ((pid = fork ()) == -1) {
      omLog (OM_LOG_CRIT, "fork after setsid");
      exit (1);
    }
    else if (pid != 0) {
      exit (0);
    }

    chdir ("/");
    umask (0);

    close (STDIN_FILENO);
    close (STDOUT_FILENO);
    close (STDERR_FILENO);
  }

  sigemptyset (&sa.sa_mask);
  sa.sa_handler = omSysSignalHandler;
  sa.sa_flags = 0;		/* XXX */

  sigaction (SIGHUP, &sa, NULL);
  sigaction (SIGTERM, &sa, NULL);
  sigaction (SIGINT, &sa, NULL);
  sigaction (SIGUSR1, &sa, NULL);
  sigaction (SIGUSR2, &sa, NULL);
  sigaction (SIGCHLD, &sa, NULL);

  sigaction (SIGALRM, &sa, NULL);
}

void
omSleep (int period)
{
  struct timespec req, rem;

  req.tv_sec = period;
  req.tv_nsec = 0;
  while (nanosleep (&req, &rem) < 0) {
    sys_errno = errno;
    if (sys_errno == EINTR) {
      req.tv_sec = rem.tv_sec;
      req.tv_nsec = rem.tv_nsec;
    }
    else {
      omLog (OM_LOG_ALERT, "omSleep:%s", strerror (sys_errno));
      break;
    }
  }
}




#if 0
sa.sa_handler = sigchld_handler;	/* reap all dead processes */
sigemptyset (&sa.sa_mask);
sa.sa_flags = SA_RESTART;

if (sigaction (SIGCHLD, &sa, NULL) == -1) {
  exit (1);
  return 1;
}

sigaction (SIGQUIT, &sa, NULL);

sigemptyset (&sa.sa_mask);
sa.sa_handler = sys_signal_stop;
sa.sa_flags = SA_RESTART;
sigaction (SIGTSTP, &sa, NULL);
sigaction (SIGCONT, &sa, NULL);

#endif
