/*
 * Copyright (c) 2005 by Nathan Wiegand
 *
 * Permission to use, copy and distribute this software is hereby granted
 * under the terms of version 2 or any later version of the GNU General Public
 * License, as published by the Free Software Foundation.
 *
 * THIS SOFTWARE IS PROVIDED IN ITS "AS IS" CONDITION, WITH NO WARRANTY
 * WHATSOEVER. NO LIABILITY OF ANY KIND FOR ANY DAMAGES WHATSOEVER RESULTING
 * FROM THE USE OF THIS SOFTWARE WILL BE ACCEPTED.
 */
/* 
 * Authors: Nathan Wiegand (nathanwiegand@gmail.com)
 * Authors: G4SRA (cradford@bcs.org.uk)
 */

#include <poll.h>
#include <arpa/telnet.h>

#include "request.h"
#include "bequest.h"
#include "tcp.h"
#include "omsyslog.h"
#include "omstatus.h"
#include "sys.h"
#include "libmisc.h"
#include "openmosix.h"
#include "node.h"

unsigned short mosix_tcp_port_nbo = 0;

/**
 * tcp_connect - Get a descriptor for a socket and connect it
 * @destaddr:	designated destination address
 * @sd:		container for descriptor of connected socket 
 * return:	valid descriptor values >= 0 or error -1
 **/
int
tcp_connect (struct sockaddr_in *destaddr, int *sd)
{
  if ((*sd = socket (destaddr->sin_family, SOCK_STREAM, 0)) != -1) {
    if (connect (*sd, (struct sockaddr *) destaddr, sizeof (struct sockaddr))
	!= -1) {
#ifdef _DEBUG_
      omLog (OM_LOG_DEBUG,
	     "tcp_connect: connected to %s:%d (%08x)",
	     inet_ntoa (destaddr->sin_addr),
	     destaddr->sin_port, ntohl (destaddr->sin_addr.s_addr));
#endif
      return (*sd);
    }
    else {
      sys_errno = errno;
      omLog (OM_LOG_INFO,
	     "tcp_connect: connect %s (%08x) failed :%s",
	     inet_ntoa (destaddr->sin_addr),
	     ntohl (destaddr->sin_addr.s_addr), strerror (sys_errno));
      tcp_send_selff (MSG_BEQ_NA_DEL "%08x\n", destaddr->sin_addr.s_addr);
    }
  }
  else {
    sys_errno = errno;
    omLog (OM_LOG_INFO, "tcp_connect: create socket failed :%s",
	   strerror (sys_errno));
  }
  return (-1);
}

/**
 * tcp_send - Send data to destination
 * @sd:		descriptor of socket connected to destination
 * @buf:	zero terminated data string buffer to be sent
 * return:	number of bytes sent >= 0 or error -1
 **/
int
tcp_send (int *sd, char *buf)
{
  int ret;

#ifdef _ROBUST_
  if (buf[strlen (buf) - 1] == '\n') {
    while (buf[strlen (buf) - 2] == '\n') {
      omLog (OM_LOG_DEBUG,
	     "ROBUST tcp_send: surplus terminating newline \"%s\"", buf);
      buf[strlen (buf) - 1] = '\0';
    }
  }
  else {
    omLog (OM_LOG_DEBUG, "ROBUST tcp_send: no terminating newline \"%s\"",
	   buf);
  }
#endif
  /*if ((ret = send (*sd, buf, strlen (buf), 0)) != -1) { */
  if ((ret = write (*sd, buf, strlen (buf))) != -1) {
#ifdef _DEBUG_
    omLog (OM_LOG_DEBUG, "tcp_send: sent \"%s\"", buf);
#endif
    return (ret);
  }
  else {
    sys_errno = errno;
    omLog (OM_LOG_CRIT, "tcp_send: error sending \"%s\" :%s", buf,
	   strerror (sys_errno));
  }
  return (-1);
}

char *
tcp_recv (int *sd)
{
  char *buf;
  int numbytes;

  buf = (char *) omMalloc (sizeof (char) * MAX_MESSAGE_SIZE + 1);
  memset (buf, '\0', MAX_MESSAGE_SIZE + 1);
  if ((numbytes = recv (*sd, buf, MAX_MESSAGE_SIZE, 0)) != -1) {
    buf[numbytes] = '\0';
#ifdef _DEBUG_
    omLog (OM_LOG_DEBUG, "tcp_recv: received \"%s\"", buf);
#endif
    return (buf);
  }
  else {
    sys_errno = errno;
    if (sys_errno != EINTR) {
      omLog (OM_LOG_CRIT, "tcp_recv: error receiving \"%s\" :%s", buf,
	     strerror (sys_errno));
    }
  }
  free (buf);
  return (NULL);
}

int
tcp_chat_start (uint32_t addr)
{
  int sockfd;
  struct sockaddr_in dest_addr;

  dest_addr.sin_addr.s_addr = addr;
  dest_addr.sin_port = mosix_tcp_port_nbo;
  dest_addr.sin_family = AF_INET;
  memset (&(dest_addr.sin_zero), '\0', 8);

  if (tcp_connect (&dest_addr, &sockfd) != -1) {
    return (sockfd);
  }
  return (-1);
}

/* send a message */
int
tcp_sendf (int *sd, const char *fmt, ...)
{
  int retn;

  va_list vl;
  char buf[MAX_MESSAGE_SIZE + 1] = { 0 };

  memset (buf, '\0', sizeof (buf));
  va_start (vl, fmt);
  retn = vsnprintf (buf, MAX_MESSAGE_SIZE, fmt, vl);
  va_end (vl);

  /* (unsigned int) To regard (-1) as greater than MAX_MESSAGE_SIZE */
  if (((unsigned int) retn) < MAX_MESSAGE_SIZE) {
    /* if (buf[strlen (buf) - 1] != '\n')
       buf[strlen (buf)] = '\n'; */
    if (tcp_send (sd, buf) != -1) {
      return (OM_STATUS_SUCCESS);
    }
  }
  else {
    omLog (OM_LOG_INFO, "tcp_sendf: failed formatting message");
  }
  return (OM_STATUS_FAIL);
}

int
tcp_send_selff (const char *fmt, ...)
{
  char buf[MAX_MESSAGE_SIZE + 1] = { 0 };
  int retn;
  va_list vl;
  int sd;

  memset (buf, '\0', sizeof (buf));
  va_start (vl, fmt);
  retn = vsnprintf (buf, MAX_MESSAGE_SIZE, fmt, vl);
  va_end (vl);

  if (((unsigned int) retn) >= MAX_MESSAGE_SIZE) {
    omLog (OM_LOG_INFO, "tcp_send_selff failed: formatting message");
    return (OM_STATUS_FAIL);
  }

  /* talk to ourself */
  if ((sd = tcp_chat_start (INADDR_ANY)) != (-1)) {
    if (tcp_send (&sd, buf) != -1) {
      close (sd);
      return (OM_STATUS_SUCCESS);
    }
    close (sd);
  }
  return (OM_STATUS_FAIL);
}

char *
tcp_exchange_msg2 (uint32_t addr, char *msg_snd)
{
  int sockfd;
  char *buf;

  if ((sockfd = tcp_chat_start (addr)) != -1) {
    if (tcp_send (&sockfd, msg_snd) != -1) {
      if ((buf = tcp_recv (&sockfd)) != NULL) {
	close (sockfd);
	return (buf);
      }
      else {
	struct in_addr sinaddr;
	sinaddr.s_addr = addr;
	omLog (OM_LOG_INFO,
	       "tcp_exchange_msg2: error receiving from %s (%08x)",
	       inet_ntoa (sinaddr), ntohl (addr));
      }
    }
    else {
      struct in_addr sinaddr;
      sinaddr.s_addr = addr;
      omLog (OM_LOG_INFO,
	     "tcp_exchange_msg2: error sending to %s (%08x)",
	     inet_ntoa (sinaddr), ntohl (addr));
    }
    close (sockfd);
  }
  return (NULL);
}

char *
tcp_exchange_msg (char *ip, char *msg_snd)
{
  struct in_addr addr;

  inet_aton (ip, &addr);
  return (tcp_exchange_msg2 (addr.s_addr, msg_snd));
}

/*
int
tcp_go_between2 (uint32_t s_addr, char *req, char *beq)
{
  char *rbuf;

  if ((rbuf = tcp_exchange_msg2 (s_addr, req)) != NULL) {
    tcp_send_selff ("%s%08x %s", beq, s_addr, rbuf);
    free (rbuf);
    return (OM_STATUS_SUCCESS);
  }
  return (OM_STATUS_FAIL);
}

int
tcp_go_between (char *ip, char *req, char *beq)
{
  struct in_addr addr;

  inet_aton (ip, &addr);
  return (tcp_go_between2 (addr.s_addr, req, beq));
}
*/

int
omd_tcp_server_init (uint32_t * addr, int *sfd)
{
  int sockfd;
  int yes = 1;
  struct sockaddr_in my_addr;

  my_addr.sin_addr.s_addr = INADDR_ANY;	/* automatically fill my IP */
  my_addr.sin_port = mosix_tcp_port_nbo;	/* short, network byte order */
  my_addr.sin_family = AF_INET;
  memset (&(my_addr.sin_zero), '\0', 8);	/* zero rest of the struct */

  /* Create a tcp socket */
  if ((sockfd = socket (AF_INET, SOCK_STREAM, 0)) == -1) {
    sys_errno = errno;
    omLog (OM_LOG_CRIT, "omd_tcp_server:socket:%s", strerror (sys_errno));
    return (OM_STATUS_FAIL);
  }

  if (setsockopt (sockfd, SOL_SOCKET, SO_REUSEADDR, &yes, sizeof (int)) == -1) {
    sys_errno = errno;
    omLog (OM_LOG_CRIT, "omd_tcp_server:setsockopt:%s", strerror (sys_errno));
    return (OM_STATUS_FAIL);
  }

  /* Name (assign a local address) to the socket */
  if (bind (sockfd, (struct sockaddr *) &my_addr, sizeof (struct sockaddr))
      == -1) {
    sys_errno = errno;
    omLog (OM_LOG_CRIT, "omd_tcp_server:bind:%s", strerror (sys_errno));
    return (OM_STATUS_FAIL);
  }

  /* Enable the socket with define queue length */
  if (listen (sockfd, BACKLOG) == -1) {
    sys_errno = errno;
    omLog (OM_LOG_CRIT, "omd_tcp_server:listen:%s", strerror (sys_errno));
    return (OM_STATUS_FAIL);
  }
  *addr = my_addr.sin_addr.s_addr;
  *sfd = sockfd;
  return (OM_STATUS_SUCCESS);
}

int
omd_tcp_server_accept (int sockfd)
{
  struct sockaddr_in their_addr;
  int new_fd;
  int ret;
  socklen_t sin_size;

  while (1) {			/* main accept() loop */
    int numbytes;
    char recv_buf[MAX_MESSAGE_SIZE + 1] = { 0 };

    if (sys_sighandled) {
      if (sys_sighandled & SYS_GOT_SIGTERM) {
	sys_sighandled &= ~SYS_GOT_SIGTERM;
	omLog (OM_LOG_INFO, "TCP Server daemon exiting on SIGTERM.");
	return (OM_STATUS_CRIT);
      }
      if (sys_sighandled & SYS_GOT_SIGINT) {
	sys_sighandled &= ~SYS_GOT_SIGINT;
	if (!(sys_flags & SYS_FLAGS_DAEMON)) {
	  /* Console mode */
	  omLog (OM_LOG_INFO, "TCP Server non-daemon exiting on SIGINT.");
	  return (OM_STATUS_CRIT);
	}
	/* Daemon mode */
      }
      if (sys_sighandled & SYS_GOT_SIGALRM) {
	sys_sighandled &= ~SYS_GOT_SIGALRM;
	omAgeNodeData ();
      }
      if (sys_sighandled & SYS_GOT_SIGHUP) {
	sys_sighandled &= ~SYS_GOT_SIGHUP;
	omLogToggleDebug ();
      }
      if (sys_sighandled & SYS_GOT_SIGUSR1) {
	sys_sighandled &= ~SYS_GOT_SIGUSR1;
	omd_print_cluster (OM_LOG_INFO);
      }
      if (sys_sighandled & SYS_GOT_SIGUSR2) {
	sys_sighandled &= ~SYS_GOT_SIGUSR2;
      }
      if (sys_sighandled & SYS_GOT_SIGCHLD) {
	sys_sighandled &= ~SYS_GOT_SIGCHLD;
/*FIXME: we need to act if the child dying is our udp half-daemon*/
	wait3 (NULL, WNOHANG, NULL);
      }
      /* NOTE: inconclusive, we could arrive here via a logic race */
      if (sys_sighandled) {
	omLog (OM_LOG_NOTICE, "Unhandled Signal %x", sys_sighandled);
	sys_sighandled = 0;
      }
    }

    sin_size = sizeof (struct sockaddr_in);
    if ((new_fd = accept (sockfd,
			  (struct sockaddr *) &their_addr,
			  &sin_size)) == -1) {
      sys_errno = errno;
      continue;
    }

/* FIXME: Deal with RFC 854 WILL/WONT DO/DONT negotiation. */

#ifdef _DEBUG_
    omLog (OM_LOG_DEBUG, "Talking With %s:%d",
	   inet_ntoa (their_addr.sin_addr), their_addr.sin_port);
#endif
    {

      /* Switch on whether or not this was a localhost asking */
      if (!strcmp ("127.0.0.1", inet_ntoa (their_addr.sin_addr))) {
	for (;;) {
	  if ((numbytes = recv (new_fd, recv_buf, MAX_MESSAGE_SIZE, 0)) == -1) {
	    sys_errno = errno;
	    if (sys_errno == EINTR) {
	      /* probably our SIGALRM whilst waiting for something to happen */
	      close (new_fd);
	      break;
	    }
	    /*FIXME: G4SRA - Why terminate with OM_STATUS_FAIL ? Where's your head at ? */
	    omLog (OM_LOG_CRIT, "recv1: %s", strerror (sys_errno));
	    close (new_fd);
	    close (sockfd);
	    return (OM_STATUS_FAIL);
	  }

	  if (numbytes <= 0)
	    break;

	  recv_buf[numbytes] = '\0';

#ifdef _DEBUG_
	  omLog (OM_LOG_DEBUG, "recv: received \"%s\"", recv_buf);
#endif
	  /* may be an internal request to modify node data */
	  {
	    char *ptr1, *ptr2;

	    /* FIXME: nasty, this modifies the buffer */
	    ret = OM_STATUS_FAIL;
	    ptr2 = recv_buf;
	    while (ptr2 != NULL) {
	      ptr1 = ptr2;
	      ptr2 = strchr (ptr1, '\n');
	      if (ptr2 != NULL) {
		*ptr2 = '\0';
		++ptr2;
		if (*ptr2 == '\0') {
		  ptr2 = NULL;
		}
	      }
	      ret = handle_bequest (ptr1);

	      if (ret < OM_STATUS_SUCCESS) {
		if (ret == OM_STATUS_BAD_BEQ) {
#ifdef _FORKED_REQUESTER_
		  if (fork () == 0) {
		    /* Child does not need this */
		    close (sockfd);
#endif
		    if ((ret =
			 handle_request (new_fd, ptr1)) < OM_STATUS_SUCCESS) {
		      omLog (OM_LOG_NOTICE, "Error [%s], local %s \"%s\"",
			     strstatus (ret),
			     inet_ntoa (their_addr.sin_addr), recv_buf);
		    }
#ifdef _FORKED_REQUESTER_
		    close (new_fd);
		    exit (0);
		  }
#endif
		}
		else {
		  omLog (OM_LOG_NOTICE,
			 "Error [%s], local %s \"%s\"",
			 strstatus (ret), inet_ntoa (their_addr.sin_addr),
			 recv_buf);
		}
	      }
	    }
	  }
	}
      }
      else {
#ifdef _FORKED_REQUESTER_
	if (fork () == 0) {
	  /* Child does not need this */
	  close (sockfd);
#endif
	  for (;;) {
	    if ((numbytes =
		 recv (new_fd, recv_buf, MAX_MESSAGE_SIZE, 0)) == -1) {
	      sys_errno = errno;
	      if (sys_errno == EINTR) {
		break;
	      }
	      omLog (OM_LOG_CRIT, "recv2: %s", strerror (sys_errno));
	      break;
	    }

	    if (numbytes <= 0)
	      break;

	    recv_buf[numbytes] = '\0';

#ifdef _DEBUG_
	    omLog (OM_LOG_DEBUG, "recv: received \"%s\"", recv_buf);
#endif
	    /* an external request for node data */
	    if ((ret = handle_request (new_fd, recv_buf)) < OM_STATUS_SUCCESS) {
	      omLog (OM_LOG_NOTICE,
		     "Error [%s], remote %s \"%s\"",
		     strstatus (ret), inet_ntoa (their_addr.sin_addr),
		     recv_buf);
	    }
	  }
#ifdef _FORKED_REQUESTER_
	  close (new_fd);
	  exit (0);
	}
#endif
      }
    }

    /* parent doesn't need this */
    close (new_fd);
  }
  close (sockfd);
  return (OM_STATUS_SUCCESS);
}
